#!/usr/bin/env python

import sys
import yaml
import math
import numpy as np
from rt import ut
from rt import mt
from rt import v
from rt import line
from rt import ax
from rt import fcx
from rt import lstx
from rt import sock
from rt import srvs

def nv_clip(nv, pa, pb):
	nv_line = line.new( mt.zero, nv )
	(in_a, in_b) = ut.map_lst( lambda p: nv_line.is_plane_side(p), [pa, pb] )
	if in_a and in_b:
		return [pa, pb]
	if not in_a and not in_b:
		return []
	pc = nv_line.cross_plane_line( line.new_p2(pa, pb) )
	return [pa, pc] if in_a else [pc, pb]

def eye_clip(wh2eye, sc_sz, pa, pb):
	(w, h, d) = sc_sz
	n = 4
	p4 = mt.arr( [ [0,0,-d], [w,0,-d], [w,h,-d], [0,h,-d] ] )
	p4 = ut.map_lst( lambda p: lstx.tr(wh2eye, 'l2g', p), p4 )
	for i in range(n):
		nv = v.p3_to_nv( mt.zero, p4[i], p4[(i+1)%n] )
		r = nv_clip(nv, pa, pb)
		if not r:
			return []
		(pa, pb) = r
	return [pa, pb]

def cvt(ps, idxs, eye2g, wh2eye, sc_sz):
	ps = ut.map_lst( lambda p: eye2g.tr('g2l', p), ps )
	p2s = ut.map_up( lambda i, j: eye_clip( wh2eye, sc_sz, ps[i], ps[j] ), idxs )
	p2s = filter( lambda p2: p2, p2s )
	p2s = map( lambda p2: ut.map_lst( lambda p: lstx.tr( wh2eye, 'g2l', p), p2 ), p2s )

	def f(p2):
		(ax, ay, _) = p2[0]
		(bx, by, _) = p2[1]
		return ( int(ax), int(ay), int(bx), int(by) )

	return list( map( f, p2s ) )

def load(fn_yaml):
	def f_read(fn):
		f = open(fn, 'r')
		s = f.read()
		f.close()
		return s
	s = sys.stdin.read() if fn_yaml == '-' else f_read(fn_yaml)
	d = yaml.load(s)

	ks = ( 'ps', 'idxs', 'eye_p', 'eye_t' )
	( ps, idxs, eye_p, eye_t ) = map( d.get, ks )
	( ps, eye_p, eye_t ) = map( mt.arr, ( ps, eye_p, eye_t ) )

	f = lambda lst: list( map( lambda i: [ lst[i], lst[i+1] ], range( len(lst) - 1 ) ) )
	idxs = sum( map( f, idxs ), [] )

	return (ps, idxs, eye_p, eye_t)

def new_wf():
	views = ( 'XYOrbit', 'TopDownOrtho' )
	e = ut.Empty()
	e.mode = ''
	e.view = views[0]

	def init(fn_yaml, w_deg):
		( e.ps, e.idxs, e.p, e.t ) = load(fn_yaml)
		e.org = ( e.p, e.t )
		e.dz = 1.0 / math.tan( 0.5 * mt.deg_to_rad(w_deg) )
	e.init = init

	def get_lst(w, h):
		sc_sz = (w, h, e.dz*w)
		wh2eye = lstx.opt( [ ax.wh_map(w, h), fcx.new(e.dz), ax.zoom_z(-1), ax.rot_x(90) ] )
		eye2g = ax.new_line( line.new_p2( e.p, e.t ) )

		if e.view == 'TopDownOrtho':
			a = v.unit( e.p - e.t )
			deg = mt.rad_to_deg( math.acos( np.dot( a, v.z1 ) ) )
			l = line.new( e.t, v.cross_product_unit( a, v.z1 ) )
			eye2g = eye2g.compo( ax.rot(l, deg) )

		return cvt(e.ps, e.idxs, eye2g, wh2eye, sc_sz)
	e.get_lst = get_lst

	def down(x, y, mode):
		e.bak = ( e.p.copy(), e.t.copy(), x, y )
		e.mode = mode
	e.down = down

	def up():
		e.mode = ''
	e.up = up

	def drag(x, y, w, h):
		(bp, bt, bx, by) = e.bak
		(w2, h2) = ( w*0.5, h*0.5 )
		(rx, ry) = ( (x - bx) / w2, (y - by) / h2 )
		if e.mode == 'zoom':
			e.p = bp + (-ry) * (bt - bp)
		elif e.mode == 'move':
			if e.view == 'TopDownOrtho':
				vy = e.t - e.p
				vy[2] = 0
				vx = ax.rot_z(-90).tr( 'l2g', vy )
				mv = ( -rx * vx + ry * vy ) * 0.1
				( e.t, e.p ) = ( bt + mv, bp + mv )
			else:
				dzw = e.dz * w
				rx = math.atan2(x - w2, dzw) - math.atan2(bx - w2, dzw)
				ry = math.atan2(y - h2, dzw) - math.atan2(by - h2, dzw)
				ax_p = ax.new_line( line.new_p2( bp, bt ) )
				e.t = lstx.tr( [ ax_p, ax.rot_z(rx*90), ax.rot_x(ry*90), ax_p.rev() ], 'g2l', bt )
		elif e.mode == 'rota':
			ax_t = ax.new_line( line.new_p2( bt, bp ) )
			e.p = lstx.tr( [ ax_t, ax.rot_z(rx*90), ax.rot_x(-ry*90), ax_t.rev() ], 'g2l', bp )
	e.drag = drag

	def reset():
		(e.p, e.t) = e.org
		e.view = views[0]
	e.reset = reset

	def turn_view():
		e.view = views[ ( views.index(e.view) + 1 ) % len(views) ]
	e.turn_view = turn_view

	e.in_drag = lambda : e.mode != ''
	e.quit = lambda : None

	return e

wf = new_wf()
cmd_infs = [
	{ 'cmd': 'init', 'func': wf.init, 'args': '(str,f8)' },
	{ 'cmd': 'get_lst', 'func': wf.get_lst, 'args': '(i4,i4)', 'rets': '[(i4,i4,i4,i4)]' },
	{ 'cmd': 'down', 'func': wf.down, 'args': '(i4,i4,str)' },
	{ 'cmd': 'up', 'func': wf.up },
	{ 'cmd': 'drag', 'func': wf.drag, 'args': '(i4,i4,i4,i4)' },
	{ 'cmd': 'reset', 'func': wf.reset },
	{ 'cmd': 'turn_view', 'func': wf.turn_view },
	{ 'cmd': 'in_drag', 'func': wf.in_drag, 'rets': 'bool' },
]

name = 'srv.wfrm'
port = 44401

def new(fn_yaml, w_deg=30, srv=None, kill_all=False):
	if srv == None:
		wf.init(fn_yaml, w_deg)
		return wf

	srvs.infs[0] = ( 'srv.val', { 'port': srvs.base_port + 11, 'cmd': 'python -m rt.val boot', 'pid': -1 } )

	if srv == 'c':
		srvs.infs += [ ( name, { 'port': port, 'cmd': './wfrm boot', 'pid': -1 } ) ]
	else:
		srvs.infs += [ ( name, { 'port': port, 'cmd': './wfrm.py boot', 'pid': -1 } ) ]

	srvs.infd = dict(srvs.infs)
	srvs.names = list( zip(*srvs.infs) )[0]
	srvs.setup()

	cli = sock.new_client(name, cmd_infs)
	cli.init(fn_yaml, w_deg)

	def quit():
		srvs.kill(name)
		if kill_all:
			srvs.kill_all()
	cli.quit = quit

	return cli

if __name__ == "__main__":
	if ut.arg_is(1, 'boot'):
		sock.server( name, cmd_infs )
		sys.exit(0)
# EOF
